# Copyright © 2024 Pavel Rabaev
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import logging
logger = logging.getLogger(__name__)

import sys

class DataObject:
    def __init__(self, pipeline, object_name):

        data_objects = pipeline.data_objects_spec
        self.pipeline_data_object_dir = pipeline.output_object_data_dpath.format(object_name=object_name)
        self.data_object_attr = self.get_object_spec(data_objects, object_name)
        if self.data_object_attr == {}:
            logger.info(f"The specification for data object {object_name} wasn't found in the section data_objects_spec[] in the pipeline: {pipeline.pipeline_fpath}")

            if pipeline.source_connector_id in ('csv'):
                logger.info(
                    f"For {pipeline.source_connector_id} extract the paramaeters data_objects[] and data_objects_spec[] are required. Use: python -m petaly init data-objects -p {pipeline.pipeline_name}")

                sys.exit()

            return self.get_default_object_spec(pipeline, object_name)

        # Source ATTRIBUTES
        self.object_name = self.data_object_attr.get('object_name')
        self.excluded_columns = self.data_object_attr.get('object_attributes').get('excluded_columns')
        self.load_type = self.data_object_attr.get('object_attributes').get('load_type')
        self.load_batch_size = self.data_object_attr.get('object_attributes').get('load_batch_size')
        self.incremental_load_column = self.data_object_attr.get('object_attributes').get('incremental_load_column')

        skip_leading_rows = self.data_object_attr.get('object_attributes').get('skip_leading_rows')
        self.skip_leading_rows = 1 if skip_leading_rows is None else skip_leading_rows

        self.file_format = self.data_object_attr.get('object_attributes').get('file_format')
        self.file_dir = self.data_object_attr.get('object_attributes').get('file_dir')
        self.file_name_list = self.data_object_attr.get('object_attributes').get('file_name_list')
        self.target_object_name =  self.data_object_attr.get('object_attributes').get('target_object_name')
        self.recreate_target_object = self.data_object_attr.get('object_attributes').get('recreate_target_object')
        if self.recreate_target_object is not True:
            self.recreate_target_object = False
        self.target_file_format = self.data_object_attr.get('object_attributes').get('target_file_format')
        self.target_file_dir = self.data_object_attr.get('object_attributes').get('target_file_dir')

    def get_object_spec(self, data_objects, object_name):

        data_object_dict = {}
        for data_object in data_objects.get('data_objects_spec'):
            if data_object != None:
                if data_object.get('object_name') == object_name:
                    data_object_dict = data_object
        return data_object_dict

    def get_default_object_spec(self, pipeline, object_name):
        self.object_name = object_name
        self.target_object_name = None
        self.recreate_target_object = False
        self.excluded_columns = None
        self.load_type = pipeline.data_object_main_config.get('preferred_load_type')
        self.column_for_incremental_load = None
        self.batch_size = None
        self.file_format = pipeline.data_object_main_config.get('data_transition_format')
        self.file_dir = None
        self.file_name_list = None
        self.skip_leading_rows = 1
        self.target_file_format = pipeline.data_object_main_config.get('data_transition_format')
        self.target_file_dir = None



# Copyright © 2024 Pavel Rabaev
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import os
import sys

from petaly.utils.file_handler import FileHandler
from petaly.core.pipeline import Pipeline
from petaly.cli.cli_menu import CliMenu


class CliInitializer():

	def __init__(self, main_config):
		self.m_conf = main_config
		self.cli_menu = CliMenu(main_config)
		self.console = self.cli_menu.console

		self.f_handler = FileHandler()

	### Init Workspace
	def init_workspace(self):
		"""
		"""
		self.create_dir(self.m_conf.pipeline_base_dpath, "Pipeline base")
		self.create_dir(self.m_conf.output_base_dpath, "Output base")
		self.create_dir(self.m_conf.logs_base_dpath, "Logs base")

	def create_dir(self, dpath, dir_type_message):

		if self.f_handler.is_dir(dpath):
			self.console.print (f"{dir_type_message} directory {dpath} already exists.")
		else:
			self.f_handler.make_dirs(dpath)
			self.console.print(f"{dir_type_message} directory {dpath} is created.")


	### Init Pipeline
	def init_pipeline(self, pipeline_name):
		""" Initiate sub folder under pipeline directory with given pipeline_name
		"""
		self.console.print(f"Init of pipeline-wizard started.")

		if pipeline_name is None:
			pipeline_name = self.cli_menu.force_assign_value(key='pipeline_name', message="Specify unique pipeline name")

		self.cli_menu.composed_pipeline_config[0]['pipeline']['pipeline_attributes'].update({'pipeline_name': pipeline_name})

		pipeline_dpath = os.path.join(self.m_conf.pipeline_base_dpath, pipeline_name)
		pipeline_fpath = os.path.join(pipeline_dpath, self.m_conf.pipeline_fname)
		output_pipeline_dpath = os.path.join(self.m_conf.output_base_dpath, pipeline_name)

		# check pipeline directory and pipeline.yaml
		if self.f_handler.is_file(pipeline_fpath):
			self.console.print (f"Pipeline with the path {pipeline_dpath} already exists.")
			process_continue = self.cli_menu.prompt.Confirm.ask(f"Do you want to continue and overwrite the existing {self.m_conf.pipeline_fname} configuration?")

			if process_continue:
				self.console.print(f"Backup with the name {self.m_conf.pipeline_fname}.buckup_<timestamp> from pipeline.yaml will be created.")
			else:
				sys.exit()

		elif self.f_handler.is_dir(pipeline_dpath):
			pass
		else:
			self.f_handler.make_dirs(pipeline_dpath)

		# check output folder and ask for confirmation
		if self.f_handler.is_dir(output_pipeline_dpath):
			self.console.print (f"Output directory with the path {output_pipeline_dpath} already exists.")
			process_continue = self.cli_menu.prompt.Confirm.ask(f"Do you want to continue and overwrite the existing output directory? All files inside will be deleted.")
			if process_continue is True:
				self.f_handler.cleanup_dir(output_pipeline_dpath)
				self.f_handler.make_dirs(output_pipeline_dpath)
			else:
				sys.exit()

		else:
			self.f_handler.make_dirs(output_pipeline_dpath)

		self.cli_menu.compose_pipeline(pipeline_name)
		self.f_handler.backup_file(pipeline_fpath)
		self.f_handler.save_dict_to_yaml(pipeline_fpath, self.cli_menu.composed_pipeline_config, dump_all=True)

		self.console.print(f"Pipeline {pipeline_name} is initialized.")
		self.console.print(f"For further configuration review the yaml file: {pipeline_fpath}")
		self.console.print(f"Output directory {output_pipeline_dpath} is created.")

		process_continue = self.cli_menu.composed_pipeline_config[0]['pipeline']['data_object_main_config'].get('use_data_objects_spec')
		if process_continue is False:
			self.console.print(f"The parameter [bold green]use_data_objects_spec[/bold green] was set to [bold blue]false[/bold blue]")
			process_continue = self.cli_menu.prompt.Confirm.ask(f"Do you want to continue defining the data objects?")

		if process_continue:
			self.cli_menu.composed_pipeline_config[0]['pipeline']['data_object_main_config'].update({'use_data_objects_spec': True})
			self.console.print(
				f"The parameter [bold green]use_data_objects_spec[/bold green] is set now to [bold blue]true[/bold blue]")
			self.init_data_objects(pipeline_name, object_names=None)
		else:
			self.console.print(
				f"Review the pipeline.yaml file and modify manually if neccessary."
				)

	def init_data_objects(self, pipeline_name, object_names):

		self.console.print(f"Initialization of data-objects started. Before running this step the pipeline {pipeline_name} should already exists.")

		if pipeline_name is None:
			pipeline_name = self.cli_menu.force_assign_value(key='pipeline_name',
													message="Provide pipeline name. Pipeline with this name should already exists.")

		pipe = Pipeline(pipeline_name, self.m_conf)

		if object_names is None:
			object_names = self.cli_menu.force_assign_value(key='object_name_list',
																message="Specify one or more comma-separated object/table names for extraction.")

		if type(object_names) == str:
			object_name_arr = [item.strip() for item in object_names.split(',')]

		pipeline_all_obj = pipe.get_pipeline_entire_config()

		data_objects_spec = self.cli_menu.compose_data_objects_spec(object_name_arr)

		if len(data_objects_spec) > 0:
			self.save_data_objects(pipeline_all_obj=pipeline_all_obj,
												 data_objects_spec=data_objects_spec,
												 pipeline_fpath=pipe.pipeline_fpath
												 )
			self.console.print(f"Data-Objects were added for pipeline {pipe.pipeline_name}. For further configuration review the yaml file: {pipe.pipeline_fpath} ")
		else:
			self.console.print(
				f"Data-Objects weren't specified for pipeline {pipe.pipeline_name}. For further configuration review the yaml file: {pipe.pipeline_fpath} ")

	def save_data_objects(self, pipeline_all_obj, data_objects_spec, pipeline_fpath):

		data_object_list = []

		# make a list of new added objects with the same index order
		for idx, obj in enumerate(data_objects_spec):
			data_object_list.insert(idx, obj.get('object_name'))

		# if object_name is in new list replace it in the pipeline, else just copy existing object on the same place
		if pipeline_all_obj[1].get('data_objects_spec') is not None:
			if len(pipeline_all_obj[1].get('data_objects_spec'))>0 and pipeline_all_obj[1].get('data_objects_spec')[0] is not None:

				for idx, data_object in enumerate(pipeline_all_obj[1].get('data_objects_spec')):

					if data_object.get('object_name') in data_object_list:
						ind, obj =  self.get_object_spec_from_array(data_objects_spec, data_object.get('object_name'))
						#pipeline_all_obj[1].get('data_objects_spec')[idx]={'object_name':data_object.get('object_name'), 'object_attributes':obj}
						pipeline_all_obj[1].get('data_objects_spec')[idx] = obj

						data_objects_spec.pop(ind)
						data_object_list.pop(ind)

		# add the entire new objects at the end of the pipeline
		if len(data_object_list) > 0:
			for i, obj in enumerate(data_objects_spec):
				pipeline_all_obj[1].get('data_objects_spec').append(obj)

		self.f_handler.backup_file(pipeline_fpath)
		self.f_handler.save_dict_to_yaml(pipeline_fpath, pipeline_all_obj, dump_all=True)

	def get_object_spec_from_array(self, object_arr, object_name):
		for idx, obj in enumerate(object_arr):
			if object_name == obj.get('object_name'):
				return idx, obj




# Copyright © 2024 Pavel Rabaev
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import os
from rich.console import Console

from petaly.utils.file_handler import FileHandler


class CliVisualizer():
	def __init__(self, main_config):
		self.m_conf = main_config
		self.file_h = FileHandler()
		self.output_line = '-------------------------------------------------------------'
		self.console = Console()


	def show_pipeline(self, pipeline_name):
		"""
		:return:
		"""
		pipeline_fpath = os.path.join(self.m_conf.pipeline_base_dpath, pipeline_name, self.m_conf.pipeline_fname)
		self.console.print(self.output_line)
		self.console.print(f"Pipeline Config {pipeline_fpath}")
		self.console.print(self.output_line)
		pipeline_config = self.file_h.load_file(pipeline_fpath)
		self.console.print(pipeline_config)
		self.console.print(self.output_line)

	def show_pipelines(self):

		file_arr = self.file_h.list_dir(path_to_dir=self.m_conf.pipeline_base_dpath)
		self.console.print(self.output_line)
		self.console.print('Pipeline Name')
		self.console.print(self.output_line)
		for i in file_arr:
			self.console.print('{0}'.format(i))

		self.console.print(self.output_line)

	def show_workspace(self):
		"""

		:return:
		"""

		if os.path.exists(self.m_conf.pipeline_base_dpath):
			self.console.print (f"Pipeline base directory: {self.m_conf.pipeline_base_dpath}")
		else:
			self.console.print(f"Missing pipeline base directory {self.m_conf.pipeline_base_dpath}. To initiate workspace run: init workspace")

		if os.path.exists(self.m_conf.output_base_dpath):
			self.console.print(f"Pipeline output directory: {self.m_conf.output_base_dpath}")
		else:
			self.console.print(f"Missing output directory {self.m_conf.output_base_dpath}. To initiate workspace run: init workspace")

		if os.path.exists(self.m_conf.logs_base_dpath):
			self.console.print(f"Logs directory: {self.m_conf.logs_base_dpath}")
		else:
			self.console.print(f"Missing logs directory {self.m_conf.logs_base_dpath}. To initiate workspace run: init workspace")




# Copyright © 2024 Pavel Rabaev
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import logging
logger = logging.getLogger(__name__)

import sys

from petaly.core.data_object import DataObject
from petaly.core.composer import Composer
from petaly.utils.file_handler import FileHandler


class ObjectMetadata():
    def __init__(self, pipeline):
        self.pipeline = pipeline
        self.f_handler = FileHandler()
        self.composer = Composer(pipeline)
        pass

    def save_table_metadata(self, meta_table):
        """ Its save a table result as a metadata into file """
        object_name = meta_table.get('source_object_name')

        source_object_fpath = self.pipeline.output_object_metadata_fpath.format(object_name=object_name)
        # format and save meta-data of each table in folder
        logger.info(f"Format and save metadata for table {object_name} in {source_object_fpath}")
        self.f_handler.save_dict_to_file(source_object_fpath, meta_table, 'json')

    def format_metadata_result(self, meta_table):
        """ Its format array of dicts of table metadata result to the array of dict with following format
        [{'table_schema': table_schema_1, 'table_name': table_name_1, 'columns': []}]
        """


        distinct_object_list = []
        formated_tbl_arr = []
        excluded_columns = []
        for i, val in enumerate(meta_table):

            schema_name = val.get("source_schema_name")
            table_name = val.get("source_object_name")

            if table_name not in distinct_object_list:
                distinct_object_list.append(table_name)
                new_dict = {'source_schema_name': schema_name, 'source_object_name': table_name, 'columns': []}
                formated_tbl_arr.append(new_dict)

        for i, val in enumerate(formated_tbl_arr):
            tbl_name = formated_tbl_arr[i].get('source_object_name')

            #data_object = self.composer.get_data_object(tbl_name)
            data_object = DataObject(self.pipeline, tbl_name)

            excluded_columns = [] if data_object.excluded_columns is None else data_object.excluded_columns

            for idx, value in enumerate(meta_table):

                table_name = meta_table[idx].get('source_object_name')
                if table_name == tbl_name:
                    if value.get('column_name') not in excluded_columns:
                        column_metadata = self.compose_column_metadata( column_name=value.get('column_name'),
                                                                    ordinal_position=value.get('ordinal_position'),
                                                                    is_nullable=value.get('is_nullable'),
                                                                    data_type=value.get('data_type'),
                                                                    character_maximum_length=value.get('character_maximum_length'),
                                                                    numeric_precision=value.get('numeric_precision'),
                                                                    numeric_scale=value.get('numeric_scale'),
                                                                    primary_key=value.get('primary_key'))

                        formated_tbl_arr[i]['columns'].append(column_metadata)

        return formated_tbl_arr

    def process_metadata(self, meta_query_result):
        object_list = []

        if meta_query_result is not None:
            for meta_table in self.format_metadata_result(meta_query_result):
                object_name = meta_table.get('source_object_name')
                object_list.append(object_name)
                self.save_table_metadata(meta_table)

        else:
            logger.error(f"Meta Query for piepline {self.pipeline.pipeline_name} is empty. Review your configuration.")
            sys.exit()

        return object_list

    def compose_column_metadata(self, column_name, ordinal_position, is_nullable, data_type, character_maximum_length, numeric_precision, numeric_scale, primary_key):
        column_meta = {}
        column_meta.update({'column_name': column_name})
        column_meta.update({'ordinal_position': ordinal_position})
        column_meta.update({'is_nullable': is_nullable})
        column_meta.update({'data_type': data_type})
        column_meta.update({'character_maximum_length': self.__replace_nan_to_none(character_maximum_length)})
        column_meta.update({'numeric_precision':  self.__replace_nan_to_none(numeric_precision)})
        column_meta.update({'numeric_scale':  self.__replace_nan_to_none(numeric_scale)})
        column_meta.update({'primary_key': primary_key})
        return column_meta

    def get_column_metadata_dict(self):
        return {
                    'column_name':None,
                    'ordinal_position': None,
                    'is_nullable': None,
                    'data_type': None,
                    'character_maximum_length': None,
                    'numeric_precision': None,
                    'numeric_scale': None,
                    'primary_key': None
                }

    def __replace_nan_to_none(self, value):

        """ This help private function. It replaces NaN or nan value to None for metadat result"""
        return_val = value if str(value).lower() != 'nan' else None
        return return_val


# Copyright © 2024 Pavel Rabaev
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


import logging
logger = logging.getLogger(__name__)

import os
import sys

from petaly.core.f_loader import FLoader


class CsvLoader(FLoader):

    def __init__(self, pipeline):
        super().__init__(pipeline)

    def load_data(self, recreate_table=False):

        object_list = super().get_data_object_list()

        for object_name in object_list:
            output_data_dir = self.pipeline.output_object_data_dpath.format(object_name=object_name)
            data_object_dict = super().get_data_object(object_name)

            # get target file directory
            target_file_dir = data_object_dict.target_file_dir
            if target_file_dir is None:
                logger.warning(f"object_attributes.target_file_dir in pipeline.yaml was not specified.")
                sys.exit()

            # get target file format
            target_file_format = data_object_dict.target_file_format
            if target_file_format is None:
                logger.warning(f"object_attributes.target_file_format in pipeline.yaml was not specified")
                target_file_format = 'csv'
                logger.info(f"File format was set to csv")

            logger.info(f"Move files from {output_data_dir} to {target_file_dir}")
            file_list = self.f_handler.get_all_dir_files(output_data_dir,
                                        target_file_format, file_names_only=True)

            for file in file_list:
                logger.info(f"File: {file}")
                file_source_fpath = os.path.join(output_data_dir, file)
                self.f_handler.cp_file(file_source_fpath, target_file_dir)



# Copyright © 2024 Pavel Rabaev
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import logging
logger = logging.getLogger(__name__)

from petaly.sysconfig.logger import setup_logging
from petaly.core.pipeline import Pipeline

class MainCtl():
    def __init__(self, main_config):
        self.m_conf = main_config
        setup_logging(self.m_conf.logging_config_fpath, self.m_conf.logs_base_dpath)

    def run_pipeline(self, pipeline, run_endpoint, object_name_list):
        """ Call this function to run pipeline source and target
        """
        #pprint(vars(pipeline))
        pipeline_name = pipeline.pipeline_name
        logger.info(f"Start: Pipeline with pipeline-name {pipeline_name} started.")
        #pipeline = Pipeline(pipeline_name, self.m_conf)

        if pipeline.is_enabled is True:
            if object_name_list is not None:
                pipeline.data_objects = object_name_list.split(',')

            if run_endpoint is None or run_endpoint == 'source':
                self.run_source(pipeline)

            if run_endpoint is None or run_endpoint == 'target':
                self.run_target(pipeline)

            logger.info(f"End: Pipeline with pipeline-name {pipeline_name} ended.")
        else:
            logger.info(f"Pipeline with pipeline-name {pipeline_name} is disabled. Check the parameter is_enabled  in pipeline.yaml file")

    ####################### run source ####################################

    def run_source(self, pipe):
        """ Call this function to run pipeline source part
        """
        logger.info("Load source config")

        class_obj = self.m_conf.get_extractor_class(pipe.source_connector_id)
        # run extraction
        if class_obj is not None:
            extractor = class_obj(pipe)
            logger.info(f"Extract connector-id: {pipe.source_connector_id}")
            extractor.extract_data()
        else:
            logger.info(f"Extractor with connector-id: {pipe.source_connector_id} can't initialized.")

    ####################### run targets ####################################

    def run_target(self, pipe):
        """ Call this function to run pipeline target part
        """
        logger.info("Load target config")

        class_obj = self.m_conf.get_loader_class(pipe.target_connector_id)

        # run loader
        if class_obj is not None:
            loader = class_obj(pipe)
            logger.info(f"Load target connector id: {pipe.target_connector_id}")
            loader.load_data()

        else:
            logger.info(f"Loader with connector-id: {pipe.target_connector_id} can't initialized.")


# Copyright © 2024 Pavel Rabaev
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import logging
logger = logging.getLogger(__name__)

import os, sys
from pyarrow import csv, parquet

from petaly.core.composer import Composer
from petaly.utils.file_handler import FileHandler
from petaly.core.object_metadata import ObjectMetadata
from petaly.core.data_object import DataObject


class FExtractor():

    def __init__(self, pipeline):
        self.pipeline = pipeline
        self.composer = Composer(pipeline)
        self.f_handler = FileHandler()
        self.object_metadata = ObjectMetadata(pipeline)
        pass

    def save_metadata_into_file(self,meta_table):
        """ Its save a table result as a metadata into file
        """
        object_name = meta_table.get('source_object_name')
        source_object_fpath = self.pipeline.output_object_metadata_fpath.format(object_name=object_name)
        logger.info(f"Format and save metadata for table {meta_table.get('source_object_name')} in {source_object_fpath}")
        self.f_handler.save_dict_to_file(source_object_fpath, meta_table, 'json')

    #def get_data_object(self, object_name):
    #    return self.composer.get_data_object(object_name)

    def extract_metadata_from_parquet_file(self, parquet_fpath):
        """ Its extract metadata from parquet file and also change the dict from parquet.read_metadata in more readable format.
        """

        pq_metadata = parquet.read_metadata(parquet_fpath).to_dict()
        pq_meta_columns = pq_metadata.get('row_groups')[0].get('columns')
        column_names = parquet.read_schema(parquet_fpath).names
        column_data_types = parquet.read_schema(parquet_fpath).types

        pq_columns_metadata_arr = []

        for i in range(len(pq_meta_columns)):
            column_dict = {}
            c_idx = column_names.index(pq_meta_columns[i].get('path_in_schema'))

            # column name
            column_name = pq_meta_columns[i].get('path_in_schema')
            column_dict.update({'column_name': column_name})

            # column type
            column_type = str(column_data_types[c_idx])
            if column_type == 'null':
                column_type = 'string'
            column_dict.update({'data_type': str(column_data_types[c_idx])})

            column_dict.update({'physical_type': str(pq_meta_columns[i].get('physical_type'))})
            column_dict.update({'is_stats_set': str(pq_meta_columns[i].get('is_stats_set'))})
            column_dict.update({'encodings': str(pq_meta_columns[i].get('encodings'))})
            column_dict.update({'compression': str(pq_meta_columns[i].get('compression'))})

            if pq_meta_columns[i].get('is_stats_set') == True:

                column_dict.update({'statistics_distinct_count': str(pq_meta_columns[i].get('statistics').get('distinct_count')) })
                column_dict.update({'statistics_min': str(pq_meta_columns[i].get('statistics').get('min'))})
                column_dict.update({'statistics_max': str(pq_meta_columns[i].get('statistics').get('max'))})
                column_dict.update({'statistics_distinct_count': str(pq_meta_columns[i].get('statistics').get('distinct_count'))})
                column_dict.update({'statistics_null_count': str(pq_meta_columns[i].get('statistics').get('null_count'))})

            else:
                column_dict.update({'statistics_distinct_count': None})
                column_dict.update({'statistics_min': None})
                column_dict.update({'statistics_max': None})
                column_dict.update({'statistics_num_values': None})
                column_dict.update({'statistics_null_count': None})


            pq_columns_metadata_arr.append(column_dict)

        return pq_columns_metadata_arr

    def compose_metadata_file(self, parquet_fpath, object_name):
        """ Its creates a metadata file with all the attributes needed to recreate a table on the target.
        """
        parquet_columns_metadata_arr = self.extract_metadata_from_parquet_file(parquet_fpath)
        if len(parquet_columns_metadata_arr) == 0:
            logger.warning("Parquet file column metadata array is empty.")
            sys.exit()

        #excluded_columns = [] if data_object.excluded_columns is None else data_object.excluded_columns

        meta_table = {}
        meta_table.update({'source_object_name': object_name})
        #data_object = self.composer.get_data_object(object_name)

        data_object = self.get_data_object(object_name)
        excluded_columns = [] if data_object.excluded_columns is None else data_object.excluded_columns

        columns_arr = []

        for idx, value in enumerate(parquet_columns_metadata_arr):

            if value.get('column_name') not in excluded_columns:

                column_metadata = self.object_metadata.compose_column_metadata(
                                                column_name=str(value.get('column_name')),
                                                ordinal_position=str(idx+1),
                                                is_nullable='YES',
                                                data_type=value.get('data_type'),
                                                character_maximum_length=None,
                                                numeric_precision=None,
                                                numeric_scale=None,
                                                primary_key=None)

                columns_arr.append(column_metadata)
        meta_table.update({'columns': columns_arr})

        return meta_table

    def print_metadata_from_pq_file(self, parquet_fpath):
        """ Its print a parquet metadata from parquet file to stdout
        """
        # extract metadata from parquet file
        pq_columns_metadata_arr = self.extract_metadata_from_parquet_file(parquet_fpath)
        header_str = ""
        col_separator = " | "
        output_text = ""

        # exit if array is empty
        if len(pq_columns_metadata_arr) == 0:
            logger.warning("pq_columns_metadata_arr is empty.")
            sys.exit()

        # create the header line
        for key in pq_columns_metadata_arr[0]:
            header_str += key + col_separator
        output_text += header_str + '\n'
        output_text += '-'*len(header_str)  + '\n'

        # print row by row
        for i in range(len(pq_columns_metadata_arr)):
            row_str = ""
            for key in pq_columns_metadata_arr[i]:
                row_str += str(pq_columns_metadata_arr[i].get(key)) + col_separator
            output_text += row_str + '\n'
        return output_text

    def analyse_file_structure(self, source_file, data_object_dict):
        """ Analyse the csv file to determine the column format.
        Transform the source file to parquet format to read the column data type.
        Finally, store the result of this discovery to the metadata file
        """
        is_file = self.f_handler.is_file(source_file)

        if is_file == False:
            logger.error(f"Source file {source_file} doesn't exists")
            sys.exit()

        if  self.f_handler.check_file_extension(source_file, data_object_dict.file_format) == False:
            logger.error(f"File format doesn't match")
            sys.exit()

        # transform csv to parquet
        file_data = csv.read_csv(source_file)
        pq_file_name = self.f_handler.replace_file_extension(source_file, '.parquet')
        parquet_fpath = os.path.join(self.pipeline.output_object_data_dpath.format(object_name=data_object_dict.object_name), pq_file_name)
        parquet.write_table(file_data, parquet_fpath)

        return parquet_fpath

    def get_data_object_list(self):
        return self.composer.get_data_object_list()

    def get_data_object(self, object_name):
        return DataObject(self.pipeline, object_name)


# Copyright © 2024 Pavel Rabaev
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from petaly.utils.file_handler import FileHandler


class Composer():

	def __init__(self, pipeline):
		self.f_handler = FileHandler(file_format='json')
		self.pipeline = pipeline
		pass


	def normalise_column_name(self, column_name):
		""" It replaces unusable characters in column name. e.g. ':' will be replaced with '_'
		"""
		column_name = column_name.replace(':', '_')
		column_name = column_name.replace('.', '_')

		return column_name

	#def get_data_object(self, object_name):
	#	return DataObject(self.pipeline, object_name)

	def _deprecated_get_data_objects(self):
		data_objects_spec = self.pipeline.data_objects_spec
		object_name_list = []

		for object in data_objects_spec:
			if object != None:
				object_name_list.append(object.get('object_name'))

		return object_name_list

	def get_data_object_list(self):
		data_objects_spec = self.pipeline.data_objects_spec
		object_name_list = []

		for data_object in data_objects_spec.get('data_objects_spec'):
			if data_object != None:
				object_name_list.append(data_object.get('object_name'))

		pipeline_data_objects = self.pipeline.data_objects
		return self.get_data_objects_intersection(object_name_list, pipeline_data_objects)

	def get_object_list_from_output_dir(self):
		object_dir_list = self.f_handler.get_all_dir_names(self.pipeline.output_pipeline_dpath)
		pipeline_object_list = self.pipeline.data_objects

		return self.get_data_objects_intersection(object_dir_list, pipeline_object_list)

	def get_data_objects_intersection(self, data_objects, pipeline_data_objects):
		""" This function check if data_object defined in pipeline.yaml data_objects corresponded with pipeline.yaml data_objects_spec.
		If pipeline.json the data_objects is None or the first element is set to None all objects from the database_schema will be loaded.
		In case it is a file export definition of objects is required

		:return:
		"""
		return_list = []

		if self.pipeline.data_object_main_config.get('use_data_objects_spec') is False:
			# return the first list without modification
			return_list = data_objects
		else:
			for value in data_objects:
				if value in pipeline_data_objects:
					return_list.append(value)

		return return_list






# Copyright © 2024 Pavel Rabaev
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import os
import logging
logger = logging.getLogger(__name__)

import sys

from petaly.core.f_extractor import FExtractor


class CsvExtractor(FExtractor):

    def __init__(self, pipeline):
        super().__init__(pipeline)

    def extract_data(self):
        """
        """
        object_list = super().get_data_object_list()

        for object_name in object_list:
            data_object = super().get_data_object(object_name)

            file_dir = data_object.file_dir
            if file_dir is None:
                logger.warning(f"object_attributes.file_dir in pipeline.yaml was not specified.")
                sys.exit()

            file_format = data_object.file_format
            if file_format is None:
                logger.warning(f"object_attributes.file_format in pipeline.yaml was not specified")
                file_format = 'csv'
                logger.info(f"File format was set to csv")


            file_list = data_object.file_name_list

            if len(file_list) == 0 or file_list[0] is None:
                file_list = self.f_handler.get_all_dir_files(file_dir, file_format, file_names_only=True)

            for file in file_list:
                file_source_fpath = os.path.join(file_dir, file)
                destination_dir = self.pipeline.output_object_data_dpath.format(object_name=object_name)
                self.f_handler.cp_file(file_source_fpath, destination_dir)

            first_file_fpath = os.path.join(destination_dir, file_list[0])

            # analyse file structure
            parquet_fpath = self.analyse_file_structure(first_file_fpath, data_object)
            meta_table = self.compose_metadata_file(parquet_fpath, object_name )
            self.save_metadata_into_file(meta_table)
            self.print_metadata_from_pq_file(parquet_fpath)


# Copyright © 2024 Pavel Rabaev
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import logging
import logging.config
import os.path
#from datetime import datetime

from petaly.utils.file_handler import FileHandler


def setup_logging(logging_config_fpath, log_dir):
    """
    """
    f_handler = FileHandler()
    config = f_handler.load_json(logging_config_fpath)
    ##date = datetime.now().strftime("%Y%m%d")
    log_fpath = os.path.join(log_dir, config.get('handlers').get('file').get('filename'))
    config['handlers']['file']['filename'] = log_fpath
    logging.config.dictConfig(config)




# Copyright © 2024 Pavel Rabaev
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import argparse
import logging
import sys
import os

#from petaly.core.logger import setup_logging

from rich.console import Console

from petaly.cli.cli_initializer import CliInitializer
from petaly.cli.cli_visualizer import CliVisualizer
from petaly.cli.cli_cleanup import CliCleanup
from petaly.core.main_ctl import MainCtl
from petaly.core.pipeline import Pipeline
from petaly.sysconfig.main_config import MainConfig


class Cli():

    def __init__(self, main_config=None):
        """
        """
        self.m_conf = MainConfig() if main_config == None else main_config
        self.console = Console()
        self.top_level_argument_message = (
                                f"Type one of the following top level positional arguments followed by optional argument: show, init, run, cleanup."
                                f"\nUse -h for help"
        )

        self.parser = argparse.ArgumentParser()
        self.parser.add_argument('top_level_argument', choices=['show', 'init', 'run', 'cleanup'], help=self.top_level_argument_message)
        self.parser.add_argument('-w', '--workspace', action="store_true", help='Provide attribute --workspace. Check exiting workspace_name by show workspace_name')
        self.parser.add_argument('-p', '--pipeline_name', help='Provide pipeline name. Check exiting pipelines by show pipelines')
        self.parser.add_argument('-o', '--object_name', help='Provide object name or a comma-separated list without empty space. The pipeline name should be specified with -p paramater too.')
        self.parser.add_argument('-c', '--config_file_path', nargs='?', type=str, help=self.m_conf.get_petaly_config_path_message())
        self.parser.add_argument('-s', '--source_only', action='store_true', help='This parameter allow to run the extract from the source only.')
        self.parser.add_argument('-t', '--target_only', action='store_true', help='This parameter allow to run the load to the target only.')
        self.parser.set_defaults(func=self.process_p)


    def process_p(self, args):
        """
        """
        if args.top_level_argument == 'show':
            self.show_p(args)
        elif args.top_level_argument == 'init':
            self.init_p(args)
        elif args.top_level_argument == 'run':
            self.run_p(args)
        elif args.top_level_argument == 'cleanup':
            self.cleanup_p(args)
        else:
            #self.exit_with_help(args.config_file_path, self.top_level_argument_message)
            pass

    def init_p(self, args):
        """
        """
        self.m_conf.set_main_config_fpath(args.config_file_path)
        self.m_conf.set_base_dpaths()
        initialize = CliInitializer(self.m_conf)

        if args.workspace:
            initialize.init_workspace()
        elif args.pipeline_name:
            if args.object_name:
                initialize.init_data_objects(args.pipeline_name, args.object_name)
            else:
                initialize.init_pipeline(args.pipeline_name)
        else:
            self.console.print('Provide one of the following sub arguments: --workspace, --pipeline_name, --pipeline_name + --object_name. Use -h for help')
            #self.parser.print_help()

    def show_p(self, args):
        """
        """
        self.m_conf.set_main_config_fpath(args.config_file_path)
        self.m_conf.set_base_dpaths()

        visualize = CliVisualizer(self.m_conf)

        if args.workspace:
            visualize.show_workspace()
        elif args.pipeline_name:
            if args.pipeline_name is None or args.pipeline_name == '':
                self.console.print("Attribute pipeline name has to be specified: --pipeline_name your_pipeline_name")
                sys.exit()
            visualize.show_pipeline(args.pipeline_name)
        else:
            visualize.show_pipelines()
            self.parser.print_help()

    def run_p(self, args):
        """
        """

        self.m_conf.set_main_config_fpath(args.config_file_path)
        self.m_conf.set_base_dpaths()

        if args.pipeline_name:

            main_ctl = MainCtl(self.m_conf)
            pipeline = Pipeline(args.pipeline_name, self.m_conf)

            if self.are_endpoints_identical(pipeline):
                self.console.print(f"In the pipeline {args.pipeline_name} source_attributes and target_attributes are exactly the same. To avoid accidentally recreating the same tables, specify at least a different schema or database name.")
                sys.exit()

            if pipeline:
                run_endpoint = None
                if args.source_only:
                    run_endpoint = 'source'
                    self.console.print(f"Run source only")
                elif args.target_only:
                    run_endpoint = 'target'
                    self.console.print(f"Run target only")
                main_ctl.run_pipeline(pipeline, run_endpoint, args.object_name)

        else:
            self.parser.print_help()
            self.console.print('Provide -p pipeline name. Check exiting pipelines below')
            sys.exit()

    def are_endpoints_identical(self, pipeline):
        """
        """
        identical_attributes = False

        source_category = self.m_conf.get_connector_class_config(pipeline.source_attr.get('endpoint_type')).get('connector_category')
        target_category = self.m_conf.get_connector_class_config(pipeline.target_attr.get('endpoint_type')).get('connector_category')

        if source_category == 'database':
            if source_category == target_category:
                for key in pipeline.source_attr.keys():
                    if key not in ('platform_type','database_password'):
                        if str(pipeline.target_attr.get(key)) == str(pipeline.source_attr.get(key)):
                            identical_attributes = True
                        else:
                            return False

        return identical_attributes

    def cleanup_p(self, args):
        """
        """
        self.m_conf.set_main_config_fpath(args.config_file_path)
        self.m_conf.set_base_dpaths()

        cleanup = CliCleanup(self.m_conf)

        if args.pipeline_name:
            if args.object_name:
                cleanup.cleanup_data_objects(args.pipeline_name, args.object_name)
            else:
                self.exit_with_help(args.config_file_path,
                                    f"Cleanup of pipeline is not supported yet. "
                                    f"To remove an object from specific pipeline provide additionaly --object_name: object_name.")
        else:
            self.exit_with_help(args.config_file_path,
                'Provide one of the following arguments: --objects. Check below if the pipeline name already exists.')

    def exit_with_help(self, config_file_path, message):
        """
        """
        self.parser.print_help()
        self.console.print('\n'+message+'\n')
        visualize = CliVisualizer(self.m_conf)
        visualize.show_pipelines()
        sys.exit()

    def start(self):
        """
        """
        try:
            args = self.parser.parse_args()
            args.func(args)
        except:

            if 'args' not in locals():
                self.console.print(
                    "----------------------------------------------------------------------------------------")
                self.console.print(self.top_level_argument_message)
                self.console.print(
                    "----------------------------------------------------------------------------------------")
                sys.exit(0)




# Copyright © 2024 Pavel Rabaev
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import logging
logger = logging.getLogger(__name__)

import os, sys
import psycopg
from psycopg.rows import dict_row


class PsqlConnector():

    def __init__(self, endpoint_attr):
        """
        """
        self.connector_id = 'postgres'
        self.column_quotes = '"'
        self.conn = self.get_connection(endpoint_attr)

    def get_connection_dsn(selfg):
        """
        """
        return os.getenv("PSQL_DSN")



    def get_connection(self, endpoint_attr):
        """
        """
        try:
            connection_string = self.compose_connection_string(endpoint_attr)

            conn = psycopg.connect(connection_string, row_factory=dict_row)

        except (Exception, psycopg.DatabaseError) as error:
            logger.error(error)
            sys.exit()

        return conn

    def compose_connection_string(self, endpoint_attr):
            """ It composes connection for Postgresql based on variables in pipeline.yaml file

            :return: return connection string
            """

            connection_string = "postgresql://{user}:{password}@{host}:{port}/{database}".format(
                                    user=endpoint_attr.get('database_user'),
                                    password=endpoint_attr.get('database_password'),
                                    host= endpoint_attr.get('database_host'),
                                    port= endpoint_attr.get('database_port'),
                                    database=endpoint_attr.get('database_name') )

            return connection_string

    def get_query_result(self, sql):
        """
        """
        try:
            with self.conn.cursor() as cur:
                cur.execute(sql)
                rows = cur.fetchall()
                logger.info(f"Found {cur.rowcount} rows in query.")
                return rows

        except (Exception, psycopg.DatabaseError) as error:
            logger.info(sql)
            logger.error(error)
            sys.exit()

    def extract_to(self, extract_to_stmt, data_fpath):
        """
        """
        with open(data_fpath, "wb") as f:
            with self.conn.cursor() as cur:
                with cur.copy(extract_to_stmt) as copy:
                    for data in copy:
                        f.write(data)

    def load_from(self, load_from_stmt, data_fpath):
        """
        """
        BLOCK_SIZE = 8192

        with open(data_fpath, "r") as f:
            try:
                with self.conn.cursor() as cur:
                    with cur.copy(load_from_stmt) as copy:
                        while data := f.read(BLOCK_SIZE):
                            copy.write(data)
                    self.conn.commit()
            except (Exception, psycopg.DatabaseError) as error:
                logger.error(error)

    def drop_table(self, schema_table_name):
        try:
            sql = f"DROP TABLE IF EXISTS {schema_table_name}"
            self.conn.execute(sql)  # Make an API request.
            self.conn.commit()
            logger.info(f"Table {schema_table_name} was dropped.")
        except (Exception, psycopg.DatabaseError) as error:
            logger.info(sql)
            logger.error(error)

    def execute_sql(self, sql):
        try:
            self.conn.execute(sql)  # Make an API request.
            self.conn.commit()
            logger.info("Query was executed.")
        except (Exception, psycopg.DatabaseError) as error:
            logger.info(sql)
            logger.error(error)


# Copyright © 2024 Pavel Rabaev
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import rich.prompt as prompt
from rich.console import Console

from petaly.utils.file_handler import FileHandler
from collections import Counter, OrderedDict


class OrderedCounter(Counter, OrderedDict):
    'Counter that remembers the order elements are first seen'
    def __repr__(self):
        return '%s(%r)' % (self.__class__.__name__,
                           OrderedDict(self))

    def __reduce__(self):
        return self.__class__, (OrderedDict(self),)

class CliMenu():

    def __init__(self, main_config):
        self.use_long_form_wizard = True
        self.console = Console()
        self.prompt = prompt
        self.m_conf = main_config
        self.f_handler = FileHandler()

        self.pipeline_meta_config = self.f_handler.load_json(self.m_conf.pipeline_meta_config_fpath)

        self.composed_pipeline_config = [
                                            {"pipeline":
                                                 {"pipeline_attributes": {},
                                                  "source_attributes": {},
                                                  "target_attributes": {},
                                                  "data_object_main_config": {}
                                                  }
                                            },
                                            {
                                              "data_objects_spec": []
                                            }
                                        ]

    def force_assign_value(self, key, message):
        while True:
            self.console.print(message)
            value = prompt.Prompt.ask(f"[bold green]{key}[/bold green]")
            if value.strip() != '':
                break
        return value

    def compose_pipeline(self, pipeline_name):
        self.use_long_form_wizard = prompt.Confirm.ask("Use long form")

        self.compose_pipeline_attributes(pipeline_name)
        self.compose_endpoint_attributes('source')
        self.compose_endpoint_attributes('target')
        self.compose_data_object_main_config()

    def compose_pipeline_attributes(self, pipeline_name):
        """
        """
        # self.pipeline_meta_config.get("pipeline_attributes")
        predefined_values = {'pipeline_name': pipeline_name}
        pipeline_attributes = self.pipeline_meta_config.get("pipeline_attributes")
        assigned_attributes = self.assign_attributes(pipeline_attributes, predefined_values=predefined_values)
        self.composed_pipeline_config[0]['pipeline']['pipeline_attributes'].update(assigned_attributes)

    def compose_endpoint_attributes(self, endpoint):
        endpoint_attributes_name = endpoint + '_attributes'

        self.console.print(
            f"---------- Specify [bold yellow]{endpoint}[/bold yellow] attributes --------------")
        endpoint_attributes = self.pipeline_meta_config.get("endpoint_attributes")

        predefined_values = {}
        # step 1. specify endpoint type
        available_connectors = self.m_conf.get_available_connectors()
        endpoint_type = prompt.Prompt.ask(f"Specify [bold yellow]{endpoint} endpoint[/bold yellow] type", choices=available_connectors)
        predefined_values.update({'endpoint_type': endpoint_type})

        assigned_endpoint_attributes = self.assign_attributes(endpoint_attributes, predefined_values=predefined_values)
        self.composed_pipeline_config[0]['pipeline'][endpoint_attributes_name].update(assigned_endpoint_attributes)

        # step 2. get connector category
        connector_category = self.m_conf.get_connector_class_config(endpoint_type).get('connector_category')

        # step 3. based on connector category define database or file parameters
        if connector_category == 'database':
            database_attributes = self.m_conf.get_database_attributes(endpoint_type)
            assigned_database_attributes = self.assign_attributes(database_attributes, predefined_values=None)
            self.composed_pipeline_config[0]['pipeline'][endpoint_attributes_name].update(assigned_database_attributes)

        elif connector_category == 'file':
            file_attributes = self.pipeline_meta_config.get("file_attributes")
            assigned_file_attributes = self.assign_attributes(file_attributes)
            self.composed_pipeline_config[0]['pipeline'][endpoint_attributes_name].update(assigned_file_attributes)

        # step 4. get and define platform type
        platform_type_list = self.m_conf.get_supported_platforms(endpoint_type)

        if len(platform_type_list) == 1:
            platform_type = platform_type_list[0]
        else:
            platform_type = prompt.Prompt.ask(f"Specify [bold green]platform_type[/bold green]",
                                              choices=platform_type_list)

        # step 5. assign platform type and platform attributes
        if platform_type != 'local':
            predefined_values.update({'platform_type': platform_type})
            platform_attributes = self.m_conf.get_platform_attributes(platform_id=platform_type)
            assigned_platform_attributes = self.assign_attributes(platform_attributes, exclude_key_list=['endpoint_type'], predefined_values=predefined_values)
            self.composed_pipeline_config[0]['pipeline'][endpoint_attributes_name].update(assigned_platform_attributes)


    def compose_data_object_main_config(self):

        self.console.print(
            f"---------- Specify [bold]Data Objects Main Config[/bold] --------------")

        data_object_main_config = self.pipeline_meta_config.get("data_object_main_config")
        assigned_data_object_main_config = self.assign_attributes(data_object_main_config, predefined_values=None, exclude_key_list=[None])
        self.composed_pipeline_config[0]['pipeline']['data_object_main_config'].update(assigned_data_object_main_config)


    def compose_data_objects_spec(self, object_names):

        self.use_long_form_wizard = prompt.Confirm.ask("Use data-objects wizard")

        data_objects_spec = self.pipeline_meta_config.get("data_objects_spec")
        data_objects_spec_list = []

        if object_names is not None and len(object_names)>0:
            for obj_name in object_names:

                self.console.print(
                    f"---------- Run initialization of Data-Objects-Spec: [bold green]{obj_name}[/bold green] --------------")

                predefined_values = {'object_name': obj_name}

                assigned_attributes = self.assign_attributes(data_objects_spec, predefined_values=predefined_values, exclude_key_list=['object_name'])
                assigned_attributes.pop('object_name')
                tmp_assigned_attributes = {}
                tmp_assigned_attributes.update({'object_name': obj_name, 'object_attributes': assigned_attributes})
                data_objects_spec_list.append(tmp_assigned_attributes)

        return data_objects_spec_list

    def assign_attributes(self, spec_attributes, predefined_values=None, exclude_key_list=None) -> dict:
        assigned_attributes = {}

        if predefined_values is not None:
            assigned_attributes.update(predefined_values)

        # cleanup dict from exclude keys
        if exclude_key_list is not None:
            for key in exclude_key_list:
                if self.f_handler.check_dict_key_exist(spec_attributes, key):
                    spec_attributes.pop(key)

        for key, value in spec_attributes.items():

            # 1. handle predefined_values
            #if predefined_values is not None and predefined_values.get(key) is not None:
            if predefined_values is not None:
                if self.f_handler.check_dict_key_exist(predefined_values, key):
                    assigned_value = predefined_values.get(key)
                    assigned_attributes.update({key: assigned_value})
                    continue

            # 2. define default and preassigned_values
            preassigned_values = value.get('preassigned_values')
            default_value = None if preassigned_values[0] is None else preassigned_values[0]
            preassigned_values = None if preassigned_values[0] is None else preassigned_values
            assigned_value = default_value

            # 3. print key comment and default value
            console_print = f"{value.get('key_comment')}"
            if default_value is not None:
                console_print += f"\nDefault value: [bold blue]{default_value}[/bold blue]"

            self.console.print(console_print)

            if key == 'database_password':
                if self.use_long_form_wizard:
                    assigned_value = prompt.Prompt.ask(f"[bold green]{key}[/bold green]", password=True)

            elif value.get('key_type') == 'Integer':
                if self.use_long_form_wizard:
                    assigned_value = prompt.IntPrompt.ask(f"[bold green]{key}[/bold green]", default=default_value,
                                                          show_default=False)

            elif value.get('key_type') == 'Array':

                if self.use_long_form_wizard:
                    assigned_value = prompt.Prompt.ask(f"[bold green]{key}[/bold green]", default=default_value,
                                                       show_default=False)

                    if type(assigned_value) == str:
                        assigned_value = [item.strip() for item in assigned_value.split(',')]

                if assigned_value is None:
                    assigned_value = [None]
            else:
                if self.use_long_form_wizard:
                    assigned_value = prompt.Prompt.ask(f"[bold green]{key}[/bold green]", choices=preassigned_values,
                                                       default=default_value, show_default=False)

                if value.get('key_type') == 'Boolean':
                    assigned_value = True if assigned_value == 'true' else False

            assigned_attributes.update({key: assigned_value})

        return assigned_attributes


# Copyright © 2024 Pavel Rabaev
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from petaly.core.composer import Composer
from petaly.utils.file_handler import FileHandler

from petaly.core.data_object import DataObject

class FLoader():

    def __init__(self, pipeline):
        self.pipeline = pipeline
        self.composer = Composer(pipeline)
        self.f_handler = FileHandler()
        pass

    def get_data_object_list(self):
        return self.composer.get_data_object_list()

    def get_data_object(self, object_name):
        return DataObject(self.pipeline, object_name)


# Copyright © 2024 Pavel Rabaev
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import logging
logger = logging.getLogger(__name__)

from petaly.connectors.postgres.psql_connector import PsqlConnector
from petaly.core.db_extractor import DBExtractor
from petaly.utils.utils import FormatDict


class PsqlExtractor(DBExtractor):
    def __init__(self, pipeline):
        self.db_connector = PsqlConnector(pipeline.source_attr)
        super().__init__(pipeline)

    def extract_data(self):
        super().extract_data()

    def get_query_result(self, meta_query):
        return self.db_connector.get_query_result(meta_query)

    def extract_to(self,extractor_obj_conf):
        output_fpath = extractor_obj_conf.get('output_fpath')
        extract_to_stmt = extractor_obj_conf.get('extract_to_stmt')
        logger.info(f"Target file: {output_fpath}")
        logger.info(f"Statement to execute: {extract_to_stmt}")

        self.db_connector.extract_to(extract_to_stmt, output_fpath)


    def compose_extract_to_stmt(self, extract_to_stmt, extractor_obj_conf) -> dict:
        """ Its save copy statement into file """
        column_list = extractor_obj_conf.get('column_list')
        schema_name = extractor_obj_conf.get('source_schema_name')
        table_name = extractor_obj_conf.get('source_object_name')

        extract_to_stmt = extract_to_stmt.format_map(
        					FormatDict(column_list=column_list, schema_name=schema_name, table_name=table_name, null_as=''))

        return extract_to_stmt



# Copyright © 2024 Pavel Rabaev
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import logging
logger = logging.getLogger(__name__)

import os
import sys

from petaly.utils.file_handler import FileHandler


class Pipeline:
    def __init__(self, pipeline_name, main_config):
        logger.info("Load main ConfigHandler")
        self.m_conf = main_config
        self.pipeline_dpath = os.path.join(self.m_conf.pipeline_base_dpath, pipeline_name)
        self.pipeline_fpath = os.path.join(self.pipeline_dpath, main_config.pipeline_fname)

        self.data_dname = 'data'
        self.metadata_dname = 'metadata'
        self.object_metadata_fname = 'object_meta.json'

        self.pipeline_type_mapping_fpath = os.path.join(self.pipeline_dpath, self.m_conf.type_mapping_fname)
        self.pipeline_extract_type_transformer_fpath = os.path.join(self.pipeline_dpath,self.m_conf.extractor_type_transformer_fname)

        self.output_pipeline_dpath = os.path.join(self.m_conf.output_base_dpath, pipeline_name)
        self.output_object_data_dpath = os.path.join(self.output_pipeline_dpath, '{object_name}', self.data_dname)
        self.output_object_metadata_dpath = os.path.join(self.output_pipeline_dpath, '{object_name}', self.metadata_dname)
        self.output_object_metadata_fpath = os.path.join(self.output_object_metadata_dpath, self.object_metadata_fname)
        self.output_extract_to_stmt_fpath = os.path.join(self.output_object_metadata_dpath, self.m_conf.extract_to_stmt_fname )
        self.output_load_from_stmt_fpath = os.path.join(self.output_object_metadata_dpath, self.m_conf.load_from_stmt_fname)
        self.output_create_table_stmt_fpath = os.path.join(self.output_object_metadata_dpath, self.m_conf.create_table_stmt_fname)

        self.pipeline_name = pipeline_name

        logger.info("Load Pipeline config")
        self.f_handler = FileHandler()

        pipeline_all_obj = self.get_pipeline_entire_config()
        pipeline_dict = pipeline_all_obj[0]

        if pipeline_dict == None:
            logger.warning(f"The pipeline: {pipeline_name} does not exists under: {self.pipeline_fpath}")
            sys.exit()

        pipeline_attr = pipeline_dict.get('pipeline').get('pipeline_attributes')
        self.source_attr = pipeline_dict.get('pipeline').get('source_attributes')
        self.target_attr = pipeline_dict.get('pipeline').get('target_attributes')

        if pipeline_attr.get('pipeline_name') != pipeline_name:
            logger.warning(f"The pass parameter for pipeline_name: {pipeline_name} does not match the pipeline_name {pipeline_attr.get('pipeline_name')} defined in the corresponding file:  {self.pipeline_fpath}")
            sys.exit()

        # PIPELINE ATTRIBUTE
        self.is_enabled = pipeline_attr.get('is_enabled')

        if self.is_enabled != True:
            logger.warning(f"The pipeline: {pipeline_name} is disabled. To enable pipeline {self.pipeline_dpath} set the parameter is_enabled: true ")
            #sys.exit()

        self.source_connector_id = self.source_attr.get('endpoint_type')
        self.target_connector_id = self.target_attr.get('endpoint_type')

        self.data_object_main_config = pipeline_dict.get('pipeline').get('data_object_main_config')
        #self.incremental_batch_size = pipeline_attr.get('incremental_batch_size')
        #self.preferred_load_type = self.data_object_main_config.get('preferred_load_type')
        #self.data_transition_format = self.data_object_main_config.get('data_transition_format')
        #self.use_data_objects_spec = self.data_object_main_config.get('use_data_objects_spec')

        self.data_objects_spec = pipeline_all_obj[1]
        if self.data_objects_spec is None:
            logger.warning(
                f"The pipeline: {pipeline_name} wasn't specify properly. It is missing following definition\n\n"
                f"---\n"
                f"data_objects_spec: []\n\n"
                f"It is recommended to reconfigure the pipeline using\n"
                f"python -m petaly init pipeline -p {pipeline_name}")
            sys.exit()

        self.data_objects = []
        if len(self.data_objects_spec) > 0:
            for obj in self.data_objects_spec.get('data_objects_spec'):
                if obj is not None:
                    self.data_objects.append(obj.get('object_name'))

    def get_pipeline_entire_config(self):
        pipeline_all_obj = self.f_handler.load_yaml_all(self.pipeline_fpath)
        return pipeline_all_obj

    def get_data_objects(self):
        data_objects = []
        if self.data_objects is not None:
            if self.data_objects[0] is not None:
                data_objects = self.data_objects

        return data_objects


# Copyright © 2024 Pavel Rabaev
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import logging
logger = logging.getLogger(__name__)

from abc import ABC, abstractmethod

from petaly.core.composer import Composer
from petaly.utils.utils import measure_time
from petaly.utils.file_handler import FileHandler
from petaly.core.type_mapping import TypeMapping
from petaly.core.object_metadata import ObjectMetadata
from petaly.core.data_object import DataObject


class DBLoader(ABC):

    def __init__(self, pipeline):
        self.pipeline = pipeline
        self.f_handler = FileHandler()
        self.composer = Composer(pipeline)
        self.m_conf = pipeline.m_conf
        self.object_metadata = ObjectMetadata(pipeline)

        self.type_mapping = TypeMapping(self.pipeline)

        if self.m_conf.set_loader_paths(self.pipeline.target_connector_id):
            self.connector_load_from_stmt_fpath = self.m_conf.connector_load_from_stmt_fpath
            self.connector_create_table_stmt_fpath = self.m_conf.connector_create_table_stmt_fpath


    @abstractmethod
    def execute_sql(self, create_table_stmt):
        pass

    @abstractmethod
    def load_from(self, object_load_conf):
        pass

    @abstractmethod
    def compose_load_from_stmt(self, data_object, loader_obj_conf):
        pass

    @abstractmethod
    def drop_table(self, loader_obj_conf: dict):
        pass

    @abstractmethod
    def create_table(self, loader_obj_conf: dict):
        pass

    @measure_time
    def load_data(self):
        """  Load data into Postgres. Recreate table if parameter recreate_table=True. """

        logger.info(f"Start the load process to the target storage: {self.pipeline.target_connector_id}.")

        # 1. get and run all objects
        object_list = self.composer.get_object_list_from_output_dir()
        for object_name in object_list:
            loader_obj_conf = {}
            loader_obj_conf.update({'object_name': object_name})

            output_metadata_object_dir = self.pipeline.output_object_metadata_dpath.format(object_name=object_name)
            loader_obj_conf.update({'output_metadata_object_dir': output_metadata_object_dir})

            output_data_object_dir = self.pipeline.output_object_data_dpath.format(object_name=object_name)
            loader_obj_conf.update({'output_data_object_dir': output_data_object_dir})

            # 2. search for metadata file and load table metadata
            metadata_file = self.pipeline.output_object_metadata_fpath.format(object_name=object_name)
            table_metadata = self.f_handler.load_file_as_dict(metadata_file, 'json')

            # 3. compose table DDL components
            data_object = DataObject(self.pipeline, object_name)
            table_ddl_dict = self.compose_table_ddl(data_object, table_metadata)
            loader_obj_conf.update({'table_ddl_dict': table_ddl_dict})

            # 4. drop and recreate table
            if data_object.recreate_target_object == True:
                self.drop_table(loader_obj_conf)

            self.create_table(loader_obj_conf)

            # 5. compose statement load_from
            output_load_from_stmt_fpath = self.pipeline.output_load_from_stmt_fpath.format(object_name=object_name)
            loader_obj_conf.update({'load_from_stmt_fpath': output_load_from_stmt_fpath})

            load_from_stmt = self.compose_load_from_stmt(data_object, loader_obj_conf)
            loader_obj_conf.update({'load_from_stmt': load_from_stmt})

            # 6. load data into table
            self.load_from(loader_obj_conf)


    #def get_data_object(self, object_name):
    #    return self.composer.get_data_object(object_name)

    def get_data_object(self, object_name):
        return DataObject(self.pipeline, object_name)

    def compose_table_ddl(self, data_object, table_metadata: dict) -> (dict):
        """ Its composes statement for create table command by using table_metadata dictionary. """

        object_name = data_object.object_name
        table_name = data_object.target_object_name

        if table_name is None:
            # table_name = table_metadata.get('source_object_name')
            table_name = object_name

        schema_name = self.pipeline.target_attr.get('database_schema')

        if schema_name is not None:
            schema_table_name = f"{schema_name}.{table_name}"
        else:
            schema_table_name = table_name

        logger.info(f"Compose DDL table: {schema_table_name}")

        table_ddl_dict = {}
        table_ddl_dict.update({'table_name': table_name})
        table_ddl_dict.update({'schema_name': schema_name})
        columns_meta_arr = table_metadata.get('columns')

        connector_create_table_stmt_fpath = self.f_handler.load_file(self.connector_create_table_stmt_fpath)
        type_mapping = self.type_mapping.get_type_mapping()
        column_list = ""
        column_datatype_list = ""
        primary_key = ''

        # loop each line
        for i, column_meta in enumerate(columns_meta_arr):

            column_name = self.db_connector.column_quotes + self.composer.normalise_column_name(column_meta.get('column_name')) + self.db_connector.column_quotes
            column_list += column_name

            column_datatype_list += column_name
            column_type = type_mapping.get(column_meta.get('data_type'))

            if column_type is None:
                logger.error(f"Type mapping doesn't exists for source-connector-id: {self.pipeline.source_connector_id}, table: {table_name}, "
                              f"column: {column_name}, data-type: {column_meta.get('data_type')}, "
                              f"target_connector_id {self.pipeline.target_connector_id}."
                              )

            column_datatype_list += " " + column_type
            mode = ' NOT NULL' if column_meta.get('is_nullable') == 'NO' else ''
            column_datatype_list += mode

            if column_meta.get('primary_key') != None:
                primary_key += column_meta.get('primary_key') + ','

            column_datatype_list += ",\n"
            column_list  += ",\n"

        if primary_key != '':
            primary_key = primary_key.rstrip(',')
            # ToDo: recreate generic primary key by adding ALTER TABLE
            primary_key = f"PRIMARY KEY ({primary_key})"
            #column_datatype_list += primary_key

        column_datatype_list = column_datatype_list.rstrip(',\n')
        column_list = column_list.rstrip(',\n')

        create_table_stmt_fpath = self.pipeline.output_create_table_stmt_fpath.format(object_name=object_name)
        table_ddl_dict.update({'column_datatype_list':column_datatype_list, 'primary_key':primary_key, 'column_list':column_list})
        table_ddl_dict.update({'create_table_stmt_fpath': create_table_stmt_fpath, 'create_table_stmt': connector_create_table_stmt_fpath})

        return table_ddl_dict

    def get_column_type_with_precision(self, column_meta, type_mapping):
        """ This functions currently is not in use. """
        column_type = type_mapping.get(column_meta.get('data_type'))
        column_type_with_precision = column_type

        if column_type in ('character varying', 'varchar'):
            max_value = '21845'

            if column_meta.get('character_maximum_length') is not None:
                column_type_with_precision = column_type + '(' + column_meta.get('character_maximum_length') + ')'
            else:
                # ToDo to find a way to get max values for each database type
                column_type_with_precision = column_type + '(' + max_value + ')'

        return column_type_with_precision

    def compose_column_list(self, table_meta):
        """ It's a sub function to compose column list """
        columns_meta_arr = table_meta.get('columns')
        columns_list = ''
        for i, column_meta in enumerate(columns_meta_arr):
            column_name = self.composer.normalise_column_name(column_meta.get('column_name'))
            columns_list += self.db_connector.column_quotes + column_name + self.db_connector.column_quotes
            columns_list += ","

        columns_list = columns_list.rstrip(',')
        return columns_list




# Copyright © 2024 Pavel Rabaev
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import logging
logger = logging.getLogger(__name__)

from petaly.connectors.mysql.mysql_connector import MysqlConnector
from petaly.core.db_loader import DBLoader
from petaly.utils.utils import FormatDict


class MysqlLoader(DBLoader):

    def __init__(self, pipeline):
        #connection_params = self.get_connection_params(pipeline)
        self.db_connector = MysqlConnector(pipeline.target_attr)
        super().__init__(pipeline)

    def load_data(self):
        super().load_data()

    def drop_table(self, table_name):
        self.db_connector.drop_table(table_name)

    def execute_sql(self, create_table_stmt):
        self.db_connector.execute_sql(create_table_stmt)

    def load_from(self, loader_obj_conf):

        object_name = loader_obj_conf.get('object_name')
        load_from_stmt = loader_obj_conf.get('load_from_stmt')
        output_data_object_dir = loader_obj_conf.get('output_data_object_dir')

        # Unzip any compressed files that may be present.
        self.f_handler.gunzip_all_files(output_data_object_dir)

        file_list = self.f_handler.get_specific_files(output_data_object_dir, '*.csv')

        for path_to_data_file in file_list:

            load_from_stmt_formated = load_from_stmt.format_map(FormatDict(path_to_data_file=path_to_data_file))
            logger.info(f"Statement to execute: {load_from_stmt_formated}")
            self.db_connector.load_from(load_from_stmt_formated)

    def compose_create_table_stmt(self, object_load_conf):

        table_ddl_dict = object_load_conf.get('table_ddl_dict')
        create_table_stmt = table_ddl_dict.get('create_table_stmt')
        table_name = table_ddl_dict.get('table_name')
        column_datatype_list = table_ddl_dict.get('column_datatype_list')

        create_table_stmt = create_table_stmt.format_map(FormatDict(table_name=table_name,
                                                                    column_datatype_list=column_datatype_list,
                                                                    partition_by='',
                                                                    cluster_by='',
                                                                    table_options='',
                                                                    alter_table_primary_or_unique_key=''))
        object_load_conf.get('table_ddl_dict').update({'create_table_stmt': create_table_stmt})

        return object_load_conf

    def compose_load_from_stmt(self, data_object, loader_obj_conf):
        """ Its compose a copy from statement """

        load_from_stmt = self.f_handler.load_file(self.connector_load_from_stmt_fpath)
        table_ddl_dict = loader_obj_conf.get('table_ddl_dict')
        table_name = table_ddl_dict.get('table_name')

        column_list = '' if table_ddl_dict.get('column_list') == None else '(' + table_ddl_dict.get('column_list') + ')'
        load_from_stmt = load_from_stmt.format_map(FormatDict(table_name=table_name,
                                                              column_list=column_list,
                                                              skip_leading_rows=data_object.skip_leading_rows))
        load_from_file_fpath = loader_obj_conf.get('load_from_stmt_fpath')

        self.f_handler.save_file(load_from_file_fpath, load_from_stmt)
        return load_from_stmt

    def drop_table(self, loader_obj_conf: dict):

        table_ddl_dict = loader_obj_conf.get('table_ddl_dict')
        table_name = table_ddl_dict.get('table_name')
        self.db_connector.drop_table(table_name)

    def create_table(self, loader_obj_conf: dict):
        table_ddl_dict = loader_obj_conf.get('table_ddl_dict')
        loader_obj_conf = self.compose_create_table_stmt(loader_obj_conf)
        self.f_handler.save_file(table_ddl_dict.get('create_table_stmt_fpath'),
                                 table_ddl_dict.get('create_table_stmt'))

        self.db_connector.execute_sql(loader_obj_conf.get('table_ddl_dict').get('create_table_stmt'))



# Copyright © 2024 Pavel Rabaev
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import os
import logging
logger = logging.getLogger(__name__)

import sys
from abc import ABC, abstractmethod

from petaly.utils.utils import measure_time
from petaly.core.composer import Composer
from petaly.utils.file_handler import FileHandler
from petaly.core.object_metadata import ObjectMetadata
from petaly.core.type_mapping import TypeMapping
from petaly.core.data_object import DataObject


class DBExtractor(ABC):

	def __init__(self, pipeline):
		super().__init__()

		self.pipeline = pipeline
		self.composer = Composer(pipeline)
		self.f_handler = FileHandler()
		self.m_conf = self.pipeline.m_conf
		self.type_mapping = TypeMapping(pipeline)
		self.object_metadata = ObjectMetadata(pipeline)

		if self.m_conf.set_extractor_paths(self.pipeline.source_connector_id):
			self.connector_extract_to_stmt_fpath = self.m_conf.connector_extract_to_stmt_fpath
			self.connector_metadata_sql_fpath = self.m_conf.connector_metadata_sql_fpath
			self.query_origin = self.f_handler.load_file(self.connector_metadata_sql_fpath)


	@abstractmethod
	def extract_to(self, extractor_obj_conf):
		pass

	@abstractmethod
	def get_query_result(self, meta_query):
		pass

	@abstractmethod
	def compose_extract_to_stmt(self, extract_to_stmt, extract_config) -> dict:
		pass

	@measure_time
	def extract_data(self):
		""" Its export data as csv into pipeline output directory.
		"""

		# 1. Start with cleanup
		self.f_handler.cleanup_dir(self.pipeline.output_pipeline_dpath)

		# 2. compose_extract_scripts
		meta_query = self.compose_meta_query()

		# 3. get meta query result
		meta_query_result = self.execute_meta_query(meta_query)

		# 4. save metadata and export scripts
		object_list = self.object_metadata.process_metadata(meta_query_result)

		# run loop for each object
		for object_name in object_list:

			# 5. get all export scripts and store data into output directory
			extractor_obj_conf = self.get_extractor_obj_conf(object_name)

			# 6. run export data
			self.extract_to(extractor_obj_conf)

	def execute_meta_query(self, meta_query):
		""" compose and execute meta query and store result in json file """
		logger.info("Execute meta-query and create extract scripts")
		if meta_query is not None:
			query_result = self.get_query_result(meta_query)

		else:
			logger.error(
				f"Meta Query for pipeline {self.pipeline.pipeline_name} can not be executed. Review your configuration.")

			query_result = None
		return query_result

	def get_extractor_obj_conf(self, object_name) ->dict:

		extractor_obj_conf = {'object_name': object_name}

		# 1. compose output data dir
		output_metadata_object_dir = self.pipeline.output_object_metadata_dpath.format(object_name=object_name)
		extractor_obj_conf.update({'output_metadata_object_dir': output_metadata_object_dir})

		# 2. get and compose metadata query
		metadata_fpath = self.pipeline.output_object_metadata_fpath.format(object_name=object_name)
		table_metadata = self.f_handler.load_file_as_dict(metadata_fpath, 'json')
		extract_queries_dict = self.compose_extract_queries(table_metadata)
		extractor_obj_conf.update(extract_queries_dict)

		# 3. load stmt_extract_to.txt and transform it in later stage
		extract_to_stmt = self.f_handler.load_file(self.connector_extract_to_stmt_fpath)
		extract_to_stmt = self.compose_extract_to_stmt(extract_to_stmt, extractor_obj_conf)
		extractor_obj_conf.update({'extract_to_stmt': extract_to_stmt})

		# 4.  save extract_to_stmt under output_extract_to_file_fpath
		output_extract_to_stmt_fpath = self.pipeline.output_extract_to_stmt_fpath.format(object_name=object_name)

		extractor_obj_conf.update({'extract_to_stmt_fpath': output_extract_to_stmt_fpath})

		self.f_handler.save_file(output_extract_to_stmt_fpath, extract_to_stmt)

		# 5. create output_file_path
		output_fpath = self.get_local_output_path(object_name)
		extractor_obj_conf.update({'output_fpath': output_fpath})

		logger.info(f"Config for data extract: {extractor_obj_conf}")
		return extractor_obj_conf

	def get_local_output_path(self, object_name):

		object_dpath = self.pipeline.output_object_data_dpath.format(object_name=object_name)

		self.f_handler.make_dirs(object_dpath)
		output_fpath = os.path.join(object_dpath, object_name + '.csv')
		return output_fpath

	def compose_meta_query(self):
		""" Its compose a meta query by using a meta query file and adding schema, tables and column definitions

		:return:
		"""
		logger.info("Compose data source meta query")

		if self.pipeline.data_object_main_config.get('use_data_objects_spec') is False:
			table_stmt = ''
		else:

			if len(self.pipeline.data_objects)==0:
				logger.warning(f"Pipeline {self.pipeline.pipeline_name} in {self.pipeline.pipeline_fpath} wasn't specified properly. If use_data_objects_spec is true the data_objects_spec: [] should has an object specification")
				sys.exit()

			table_stmt = 'AND tb.table_name IN ({tbl_list})'
			table_string = ''
			for tbl in self.pipeline.data_objects:
				table_string += "'" + tbl + "',"

			table_string = table_string.rstrip(',')

			table_stmt = table_stmt.format(tbl_list=table_string)

		source_schema = self.pipeline.source_attr.get('database_schema')

		if source_schema is None:
			if self.f_handler.check_dict_key_exist(self.pipeline.source_attr, 'database_schema'):
				logger.warning(f"A source database schema wasn't specified. To continue, specify database_schema in pypeline.yaml ")
				sys.exit()

			source_schema = self.pipeline.source_attr.get('database_name')

		meta_query = self.query_origin.format(schema=source_schema, table_statement_list=table_stmt)

		return meta_query

	@measure_time
	def compose_extract_queries(self, dict_obj):
		""" Its compose an extract queries using object dictionary
		"""
		extract_obj_conf = {}

		if dict_obj.get('source_object_name') is not None:
			column_list = ''

			transformation = self.type_mapping.get_extractor_type_transformer()

			col_obj = dict_obj.get('columns')

			for idx, val in enumerate(col_obj):

				column_transformation = transformation.get(val['data_type'])

				if column_transformation is not None:
					column_list += column_transformation.format(column_name=val['column_name']) + ","
				else:
					column_list += f" {self.db_connector.column_quotes}{val['column_name']}{self.db_connector.column_quotes},"
					#columns += self.normalise_column_name(val['column_name']) + ","

			column_list = column_list.rstrip(',')

			extract_obj_conf.update({
								 'source_schema_name':dict_obj['source_schema_name'],
								 'source_object_name':dict_obj['source_object_name'],
								 'column_list':column_list})

			return extract_obj_conf

	def get_data_object(self, object_name):
		return DataObject(self.pipeline, object_name)
